/**
   http://www.infoq.com/articles/forkjoin-to-parallel-streams
*/
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveTask;
import java.util.Random;

interface SequentialCalculator
{
   double computeSequentially(double[] numbers, int start, int end);
}


class ForkJoinCalculator extends RecursiveTask<Double>
{
   public static final long THRESHOLD = 1_000_000;

   private final SequentialCalculator sequentialCalculator;
   private final double[] numbers;
   private final int start;
   private final int end;

   public ForkJoinCalculator(double[] numbers, SequentialCalculator sequentialCalculator)
   {
      this(numbers, 0, numbers.length, sequentialCalculator);
   }

   private ForkJoinCalculator(double[] numbers, int start, int end, SequentialCalculator sequentialCalculator)
   {
      this.numbers = numbers;
      this.start = start;
      this.end = end;
      this.sequentialCalculator = sequentialCalculator;
   }

   @Override
   protected Double compute()
   {
      int length = end - start;
      if (length <= THRESHOLD)
      {
         return sequentialCalculator.computeSequentially(numbers, start, end);
      }

      ForkJoinCalculator leftTask = new ForkJoinCalculator(numbers, start, start + length/2, sequentialCalculator);
      ForkJoinCalculator rightTask = new ForkJoinCalculator(numbers, start + length/2, end, sequentialCalculator);

      leftTask.fork();
      Double rightResult = rightTask.compute();
      Double leftResult = leftTask.join();
      return leftResult + rightResult;
   }
}//ForkJoinCalculator


public class Variance
{
   public static double varianceImperative(double[] population)
   {
      double average = 0.0;

      for(double p: population)
      {
         average += p;
      }

      average /= population.length;

      double variance = 0.0;

      for(double p: population)
      {
         variance += (p - average) * (p - average);
      }

      return variance/population.length;
   }//varianceImperative()



   public static double varianceForkJoin(double[] population)
   {
      final ForkJoinPool forkJoinPool = new ForkJoinPool();

      double total = forkJoinPool.invoke(
         new ForkJoinCalculator(
            population,
            new SequentialCalculator()
            {
               @Override
               public double computeSequentially(double[] numbers, int start, int end)
               {
                  double total = 0;
                  for (int i = start; i < end; i++)
                  {
                     total += numbers[i];
                  }
                  return total;
               }
            }));

      final double average = total / population.length;

      double variance = forkJoinPool.invoke(
         new ForkJoinCalculator(
            population,
            new SequentialCalculator()
            {
               @Override
               public double computeSequentially(double[] numbers, int start, int end)
               {
                  double variance = 0;
                  for (int i = start; i < end; i++)
                  {
                     variance += (numbers[i] - average) * (numbers[i] - average);
                  }
                  return variance;
               }
            }));

      return variance / population.length;
   }//varianceForkJoin()


/*
   // This is Java 8 code.
   public static double varianceStreams(double[] population)
   {
      double average = Arrays.stream(population).parallel().average().orElse(0.0);

      double variance = Arrays.stream(population).parallel().map(p->(p-average)*(p-average))
                                                            .sum() / population.length;
      return variance;
   }
*/


   public static final int PROBLEM_SIZE = 100_000_000;
   public static double numbers[] = new double[PROBLEM_SIZE];


   public static void main(String[] args)
   {
      Random generator = new Random();
      for (int i = 0; i < numbers.length; i++)
      {
         // http://docs.oracle.com/javase/6/docs/api/java/util/Random.html#nextDouble()
         //numbers[i] = generator.nextDouble();
         // http://docs.oracle.com/javase/6/docs/api/java/util/Random.html#nextGaussian()
         numbers[i] = generator.nextGaussian();
      }

      long startTime;
      long stopTime;

      // Run the single-threaded sequential version.
      startTime = System.currentTimeMillis();
      double resultImparative = varianceImperative( numbers );
      stopTime = System.currentTimeMillis();
      System.out.println("Sequential time: " + (stopTime - startTime));
      System.out.println("Sequential answer: " + resultImparative);

      // Run the fork-join version.
      startTime = System.currentTimeMillis();
      double resultForkJoin = varianceForkJoin( numbers );
      stopTime = System.currentTimeMillis();
      System.out.println(" Fork-Join time: " + (stopTime - startTime));
      System.out.println(" Fork-Join answer: " + resultForkJoin);
   }//main()
}//Variance